# bassed on: https://github.com/songlab-cal/tape/blob/master/tape/models/modeling_trrosetta.py
class DilatedResidualBlock(nn.Module):

    def __init__(self, num_features: int, kernel_size: int, dilation: int, dropout: float):
        super().__init__()
        padding = self._get_padding(kernel_size, dilation)
        self.conv1 = nn.Conv2d(
            num_features, num_features, kernel_size, padding=padding, dilation=dilation, bias=False)
        self.norm1 = nn.InstanceNorm2d(num_features, affine=True, eps=1e-6)
        self.actv1 = nn.ELU(inplace=True)
        self.dropout = nn.Dropout(dropout)
        self.conv2 = nn.Conv2d(
            num_features, num_features, kernel_size, padding=padding, dilation=dilation, bias=False)
        self.norm2 = nn.InstanceNorm2d(num_features, affine=True, eps=1e-6)
        self.actv2 = nn.ELU(inplace=True)
        self.apply(self._init_weights)
        nn.init.constant_(self.norm2.weight, 0)

    def _get_padding(self, kernel_size: int, dilation: int) -> int:
        return (kernel_size + (kernel_size - 1) * (dilation - 1) - 1) // 2

    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, nn.Conv2d):
            nn.init.kaiming_normal_(module.weight, mode='fan_out', nonlinearity='relu')
            if module.bias is not None:
                module.bias.data.zero_()

        # elif isinstance(module, DilatedResidualBlock):
            # nn.init.constant_(module.norm2.weight, 0)

    def forward(self, features):
        shortcut = features
        features = self.conv1(features)
        features = self.norm1(features)
        features = self.actv1(features)
        features = self.dropout(features)
        features = self.conv2(features)
        features = self.norm2(features)
        features = self.actv2(features + shortcut)
        return features
    
class TRRosettaPredictor(nn.Module):

    def __init__(self):
        super().__init__()
        layers = [
            nn.Conv2d(1+33*2+1280*2+660*2, 128, 1, bias=False), #seq_sep + one-hot+1D BERT tiled + 2D BERT + transposed
            nn.InstanceNorm2d(128, affine=True, eps=1e-6),
            nn.ELU()]
        
        #block1
        for _ in range(1):
            for dilation in [1,2,4,8,16]:
                block = DilatedResidualBlock(
                    128, 3, dilation, 0.15)
                layers.append(block)
        
        
        layers.append(nn.Conv2d(128, 64, 1, bias=True))
        layers.append(nn.ELU())
        
        
        #block2
        for _ in range(6):
            for dilation in [1,2,4,8,16]:
                block = DilatedResidualBlock(
                    64, 3, dilation, 0.15)
                layers.append(block)
        
        #one more block with dilation=1
        dilation=1
        block = DilatedResidualBlock(
                    64, 3, dilation, 0.15)
        layers.append(block)

        self.resnet = nn.Sequential(*layers)
        self.predict_theta = nn.Conv2d(64, 37, 1)
        self.predict_phi = nn.Conv2d(64, 19, 1)
        self.predict_dist = nn.Conv2d(64, 37, 1)
        self.predict_bb = nn.Conv2d(64, 3, 1)
        self.predict_omega = nn.Conv2d(64, 37, 1)

        self.init_weights()


    def _init_weights(self, module):
      """ Initialize the weights """
      if isinstance(module, nn.Linear):
          module.weight.data.normal_(mean=0.0, std=0.02)
          if module.bias is not None:
              module.bias.data.zero_()
      elif isinstance(module, nn.Conv2d):
          nn.init.kaiming_normal_(module.weight, mode='fan_out', nonlinearity='relu')
          if module.bias is not None:
              module.bias.data.zero_()
      elif isinstance(module, DilatedResidualBlock):
          nn.init.constant_(module.norm2.weight, 0)

    def init_weights(self):
        self.apply(self._init_weights)
        nn.init.constant_(self.predict_theta.weight, 0)
        nn.init.constant_(self.predict_phi.weight, 0)
        nn.init.constant_(self.predict_dist.weight, 0)
        nn.init.constant_(self.predict_bb.weight, 0)
        nn.init.constant_(self.predict_omega.weight, 0)

    def forward(self,
                features,
                theta=None,
                phi=None,
                dist=None,
                omega=None):
        batch_size = features.size(0)
        seqlen = features.size(2)
        embedding = self.resnet(features)

        # anglegrams for theta
        logits_theta = self.predict_theta(embedding)

        # anglegrams for phi
        logits_phi = self.predict_phi(embedding)

        # symmetrize
        sym_embedding = 0.5 * (embedding + embedding.transpose(-1, -2))

        # distograms
        logits_dist = self.predict_dist(sym_embedding)

        # beta-strand pairings (not used)
        # logits_bb = self.predict_bb(sym_embedding)

        # anglegrams for omega
        logits_omega = self.predict_omega(sym_embedding)

        logits_dist = logits_dist.permute(0, 2, 3, 1).contiguous()
        logits_theta = logits_theta.permute(0, 2, 3, 1).contiguous()
        logits_omega = logits_omega.permute(0, 2, 3, 1).contiguous()
        logits_phi = logits_phi.permute(0, 2, 3, 1).contiguous()

        probs = {}
        probs['p_dist'] = nn.Softmax(-1)(logits_dist)
        probs['p_theta'] = nn.Softmax(-1)(logits_theta)
        probs['p_omega'] = nn.Softmax(-1)(logits_omega)
        probs['p_phi'] = nn.Softmax(-1)(logits_phi)
        outputs = (probs,)

        metrics = {}
        total_loss = 0

        if dist is not None:
            logits_dist = logits_dist.reshape(batch_size * seqlen * seqlen, 37)
            loss_dist = nn.CrossEntropyLoss()(logits_dist, dist.view(-1))
            metrics['dist'] = loss_dist
            total_loss += loss_dist
        if theta is not None:
            logits_theta = logits_theta.reshape(batch_size * seqlen * seqlen, 37)
            loss_theta = nn.CrossEntropyLoss()(logits_theta, theta.view(-1))
            metrics['theta'] = loss_theta
            total_loss += loss_theta
        if omega is not None:
            logits_omega = logits_omega.reshape(batch_size * seqlen * seqlen, 37)
            loss_omega = nn.CrossEntropyLoss()(logits_omega, omega.view(-1))
            metrics['omega'] = loss_omega
            total_loss += loss_omega
        if phi is not None:
            logits_phi = logits_phi.reshape(batch_size * seqlen * seqlen, 19)
            loss_phi = nn.CrossEntropyLoss()(logits_phi, phi.view(-1))
            metrics['phi'] = loss_phi
            total_loss += loss_phi

        if len(metrics) > 0:
            outputs = ((total_loss, metrics),) + outputs

        return outputs