# Adopted from Roshan Rao (roshan_rao@berkeley.edu),
# TAPE repo https://github.com/songlab-cal/tape
# https://github.com/songlab-cal/tape/blob/master/tape/models/modeling_trrosetta.py

import os,sys
import numpy as np
import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint_sequential

#script_dir = os.path.dirname(__file__)
#sys.path.insert(0, script_dir)

from kinematics import *

class MSAFeatureExtractor(nn.Module):

    def __init__(self, wmin=0.8, dcareg=4.5):
        super(MSAFeatureExtractor, self).__init__()
        self.wmin = wmin
        self.dcareg = dcareg

    def forward(self, msa, idx, msa1hot=None):

        if msa1hot == None:
            msa1hot = torch.nn.functional.one_hot(msa, 21).float()
        seqlen = msa1hot.size(2)

        weights = self.reweight(msa1hot)
        features_1d = self.extract_features_1d(msa1hot, weights)
        features_2d = self.extract_features_2d(msa1hot, idx, weights)
        
        left = features_1d.unsqueeze(2).repeat(1, 1, seqlen, 1)
        right = features_1d.unsqueeze(1).repeat(1, seqlen, 1, 1)
        features = torch.cat((left, right, features_2d), -1)
        features = features.type(torch.float32)
        features = features.permute(0, 3, 1, 2)
        features = features.contiguous()
        return features

    def reweight(self, msa1hot, eps=1e-9):
        # Reweight
        seqlen = msa1hot.size(2)
        id_min = seqlen * self.wmin
        id_mtx = torch.stack([torch.tensordot(el, el, [[1, 2], [1, 2]]) for el in msa1hot], 0)
        id_mask = id_mtx > id_min
        weights = 1.0 / (id_mask.type_as(msa1hot).sum(-1) + eps)
        return weights

    def extract_features_1d(self, msa1hot, weights):
        # 1D Features
        f1d_seq = msa1hot[:, 0, :, :20]
        batch_size = msa1hot.size(0)
        seqlen = msa1hot.size(2)

        # msa2pssm
        beff = weights.sum()
        msw = weights[:, :, None, None] * msa1hot
        f_i = msw.sum(1) / beff + 1e-9
        h_i = (-f_i * f_i.log()).sum(2, keepdims=True)
        n_i = torch.log(msw[...,:20].sum(dim=(1,3))).unsqueeze(-1)
        f1d_pssm = torch.cat((f_i, h_i, n_i), dim=-1)
        f1d = torch.cat((f1d_seq, f1d_pssm), dim=-1)
        f1d = f1d.view(batch_size, seqlen, 43)
        return f1d

    def extract_features_2d(self, msa1hot, idx, weights):

        B,N,L,C = msa1hot.shape
        device = msa1hot.device

        if N == 1:
            # if no alignments, predict from sequence alone
            return torch.cat([torch.eye(L, device=device)[None,:,:,None].expand(B,L,L,441),
                              torch.zeros(B,L,L,4, device=device)],dim=-1)

        # compute fast_dca
        # covariance
        x = msa1hot.view(B, N, L*C)
        num_points = weights.sum(1) - weights.mean(1).sqrt()
        mean = (x * weights.unsqueeze(2)).sum(1, keepdims=True) / num_points[:, None, None]
        x = (x - mean) * weights[:, :, None].sqrt()
        cov = torch.matmul(x.transpose(-1, -2), x) / num_points[:, None, None]

        # inverse covariance
        reg = torch.eye(L*C,
                        device=device,
                        dtype=weights.dtype)[None]
        reg = reg * self.dcareg / weights.sum(1, keepdims=True).sqrt().unsqueeze(2)
        cov_reg = cov + reg
        inv_cov = torch.stack([torch.inverse(cr) for cr in cov_reg.unbind(0)], 0)

        x1 = inv_cov.view(B, L, C, L, C)
        x2 = x1.permute(0, 1, 3, 2, 4)
        features = x2.reshape(B, L, L, C*C)

        x3 = (x1[:, :, :-1, :, :-1] ** 2).sum((2, 4)).sqrt() * (
            1 - torch.eye(L, device=device, dtype=weights.dtype)[None])
        apc = x3.sum(1, keepdims=True) * x3.sum(2, keepdims=True) / x3.sum(
            (1, 2), keepdims=True)
        contacts = (x3 - apc) * (1 - torch.eye(
            L, device=device, dtype=x3.dtype).unsqueeze(0))

        # some more 2d features
        f_ij = torch.stack([torch.tensordot(el, el, [[0], [0]]) 
                            for el in msa1hot], 0) / N + 1e-9
        gaps = f_ij[:,:,20,:,20].unsqueeze(-1)
        h_ij = torch.sum( -f_ij * torch.log(f_ij), dim=(2,4)).unsqueeze(-1)
        
        seqsep = torch.abs(idx[:,:,None]-idx[:,None,:])+1.0
        seqsep = torch.log(seqsep).unsqueeze(-1)
        
        f2d_dca = torch.cat([features, contacts[...,None], gaps, h_ij, seqsep], axis=3)
        return f2d_dca


class DilatedResidualBlock(nn.Module):

    def __init__(self, num_features: int, kernel_size: int, dilation: int, dropout: float):
        super().__init__()
        padding = self._get_padding(kernel_size, dilation)
        self.conv1 = nn.Conv2d(
            num_features, num_features, kernel_size, padding=padding, dilation=dilation, bias=False)
        self.norm1 = nn.InstanceNorm2d(num_features, affine=True, eps=1e-6)
        self.actv1 = nn.ELU(inplace=True)
        self.dropout = nn.Dropout(dropout)
        self.conv2 = nn.Conv2d(
            num_features, num_features, kernel_size, padding=padding, dilation=dilation, bias=False)
        self.norm2 = nn.InstanceNorm2d(num_features, affine=True, eps=1e-6)
        self.actv2 = nn.ELU(inplace=True)

    def _get_padding(self, kernel_size: int, dilation: int) -> int:
        return (kernel_size + (kernel_size - 1) * (dilation - 1) - 1) // 2

    def forward(self, features):
        shortcut = features
        features = self.conv1(features)
        features = self.norm1(features)
        features = self.actv1(features)
        features = self.dropout(features)
        features = self.conv2(features)
        features = self.norm2(features)
        features = self.actv2(features + shortcut)
        return features


# original resblock
class ResBlock2D(nn.Module):
    def __init__(self, n_c, kernel=3, dilation=1, p_drop=0.15):
        super(ResBlock2D, self).__init__()
        padding = self._get_same_padding(kernel, dilation)

        layer_s = list()
        layer_s.append(nn.Conv2d(n_c, n_c, kernel, padding=padding, dilation=dilation, bias=False))
        layer_s.append(nn.InstanceNorm2d(n_c, affine=True, eps=1e-6))
        layer_s.append(nn.ELU(inplace=True))
        # dropout
        layer_s.append(nn.Dropout(p_drop))
        # convolution
        layer_s.append(nn.Conv2d(n_c, n_c, kernel, dilation=dilation, padding=padding, bias=False))
        layer_s.append(nn.InstanceNorm2d(n_c, affine=True, eps=1e-6))
        self.layer = nn.Sequential(*layer_s)
        self.final_activation = nn.ELU(inplace=True)

    def _get_same_padding(self, kernel, dilation):
        return (kernel + (kernel - 1) * (dilation - 1) - 1) // 2

    def forward(self, x):
        out = self.layer(x)
        return self.final_activation(x + out)

# pre-activation bottleneck resblock (AF1-style)
class ResBlock2D_bottleneck(nn.Module):
    def __init__(self, n_c, kernel=3, dilation=1, p_drop=0.15):
        super(ResBlock2D_bottleneck, self).__init__()
        padding = self._get_same_padding(kernel, dilation)

        n_b = n_c // 2 # bottleneck channel
        
        layer_s = list()
        # pre-activation
        layer_s.append(nn.InstanceNorm2d(n_c, affine=True, eps=1e-6))
        layer_s.append(nn.ELU(inplace=True))
        # project down to n_b
        layer_s.append(nn.Conv2d(n_c, n_b, 1, bias=False))
        layer_s.append(nn.InstanceNorm2d(n_b, affine=True, eps=1e-6))
        layer_s.append(nn.ELU(inplace=True))
        # convolution
        layer_s.append(nn.Conv2d(n_b, n_b, kernel, dilation=dilation, padding=padding, bias=False))
        layer_s.append(nn.InstanceNorm2d(n_b, affine=True, eps=1e-6))
        layer_s.append(nn.ELU(inplace=True))
        # dropout
        layer_s.append(nn.Dropout(p_drop))
        # project up
        layer_s.append(nn.Conv2d(n_b, n_c, 1, bias=False))

        self.layer = nn.Sequential(*layer_s)

    def _get_same_padding(self, kernel, dilation):
        return (kernel + (kernel - 1) * (dilation - 1) - 1) // 2

    def forward(self, x):
        out = self.layer(x)
        return x + out

class TRRosetta2(nn.Module):
    """trRosetta base network"""
    
    def __init__(self, 
                 wmin=0.8, dcareg=4.5, gradchk=2,
                 block_type='orig', 
                 dropout=0.15, drates=[1,2,4,8,16], fsize=3, 
                 cycls1=1, cycls2=6, filts1=128, filts2=64, 
                 dbins=36, abins=36 ):
        super(TRRosetta2, self).__init__()
        self.gradchk = gradchk

        if block_type == 'orig':
            self.resblock = ResBlock2D
        elif block_type == 'bottleneck':
            self.resblock = ResBlock2D_bottleneck
        else:
            sys.exit('Wrong residual block type: %s'%(block_type))
            
        self.feat_extractor = MSAFeatureExtractor(wmin=wmin, dcareg=dcareg)

        layers = [
            nn.Conv2d(531, filts1, 1),
            nn.InstanceNorm2d(filts1, affine=True),
            nn.ELU(),
            nn.Dropout(dropout)
        ]

        # first cycle (with more filters)
        for _ in range(cycls1):
            for dilation in drates:
                layers.append(self.resblock(filts1, fsize, dilation, dropout))

        # project down
        layers.append(nn.Conv2d(filts1, filts2, 1))
        layers.append(nn.ELU())

        # second cycle (with less filters)
        for _ in range(cycls2):
            for dilation in drates:
                layers.append(self.resblock(filts2, fsize, dilation, dropout))

        # one more block with dilation 1
        layers.append(self.resblock(filts2, fsize, 1, dropout))

        # put all together
        self.resnet = nn.Sequential(*layers)

        # branches to predict distances and orientations
        self.predict_dist = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, dbins+1, 1)
        )

        self.predict_omega = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, abins+1, 1)
        )
        
        self.predict_theta = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins+1, 1)
        )
        
        self.predict_phi = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins//2+1, 1)
        )

        self.predict_mu = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, 1, 1)
        )
        self.predict_sigma = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, 1, 1),
            nn.ELU()
        )


    def forward(self,msa,idx=None, msa_one_hot=None):

        B, N, L = msa.shape
        if idx == None:
            idx = torch.arange(L, device=msa.device).unsqueeze(0).expand(B,-1)
        
        if msa_one_hot != None:
            features = self.feat_extractor(msa,idx,msa1hot=msa_one_hot)
        else:
            features = self.feat_extractor(msa,idx)
        
        if self.gradchk > 1:
            features.requires_grad = True 
            emb = checkpoint_sequential(self.resnet, self.gradchk, features)
        else:
            emb = self.resnet(features)

        embT = torch.transpose(emb,2,3)
        
        # logits for 6D coordinate predictions
        logits = [self.predict_dist(0.5*(emb+embT)),
                  self.predict_omega(0.5*(emb+embT)),
                  self.predict_theta(torch.cat([emb,embT],dim=1)),
                  self.predict_phi(torch.cat([emb,embT],dim=1))]

        # predicted parameters of lognormal distributions
        # for Ca-Ca distances (mu,sigma)
        lognorm = torch.stack([self.predict_mu(0.5*(emb+torch.transpose(emb,2,3))).squeeze(1),
                               self.predict_sigma(0.5*(emb+torch.transpose(emb,2,3))).squeeze(1)+1],dim=-1)

        out = {"dist": logits[0], "omega": logits[1], 
               "theta": logits[2], "phi": logits[3],
               "edge": emb, "lognorm": lognorm}
    
        return out

'''
class TRRosetta2tbm(nn.Module):
    
    def __init__(self, 
                 wmin=0.8, dcareg=4.5, gradchk=2,
                 block_type='orig', 
                 dropout=0.15, drates=[1,2,4,8,16], fsize=3, 
                 cycls1=1, cycls2=6, filts1=128, filts2=64, 
                 dbins=36, abins=36 ):
        super(TRRosetta2tbm, self).__init__()
        self.params = params
        self.tplt_prepro = nn.Sequential(
            nn.Dropout(self.params['DROPOUT']),
            nn.Conv2d(ntpltfeat, self.params['TPLTEMB'], 3, padding=1),
            nn.ELU()
        )
        self.trrosetta = TRRosetta2(nmsafeat+params['TPLTEMB'], params)


    def forward(self,features,templates):

        tpltemb = self.tplt_prepro(templates).mean(0,keepdim=True)
        out = self.trrosetta(torch.cat([tpltemb,features],dim=1))
        return out
'''


class TRRosetta2tbm(nn.Module):
    """template-based trRosetta network"""
    
    def __init__(self, 
                 wmin=0.8, dcareg=4.5, gradchk=0,
                 tmpltemb=64, 
                 block_type='orig', 
                 dropout=0.15, drates=[1,2,4,8,16], fsize=3, 
                 cycls1=1, cycls2=6, filts1=128, filts2=64, 
                 dbins=36, abins=36 ):
        super(TRRosetta2tbm, self).__init__()
        self.gradchk = gradchk
        self.dbins = dbins

        if block_type == 'orig':
            self.resblock = ResBlock2D
        elif block_type == 'bottleneck':
            self.resblock = ResBlock2D_bottleneck
        else:
            sys.exit('Wrong residual block type: %s'%(block_type))
            
        self.msa_feat_extractor = MSAFeatureExtractor(wmin=wmin, dcareg=dcareg)

        self.tplt_feat_extractor = nn.Sequential(
            #nn.Dropout(dropout),
            nn.Conv2d(49, tmpltemb, 3, padding=1),
            nn.ELU()
            #nn.Conv2d(49, tmpltemb, 1),
            #nn.ELU(),
            #self.resblock(tmpltemb, 3, 1, dropout),
        )

        layers = [
            nn.Conv2d(531+tmpltemb, filts1, 1),
            nn.InstanceNorm2d(filts1, affine=True),
            nn.ELU(),
            nn.Dropout(dropout)
        ]

        # first cycle (with more filters)
        for _ in range(cycls1):
            for dilation in drates:
                layers.append(self.resblock(filts1, fsize, dilation, dropout))

        # project down
        layers.append(nn.Conv2d(filts1, filts2, 1))
        layers.append(nn.ELU())

        # second cycle (with less filters)
        for _ in range(cycls2):
            for dilation in drates:
                layers.append(self.resblock(filts2, fsize, dilation, dropout))

        # one more block with dilation 1
        layers.append(self.resblock(filts2, fsize, 1, dropout))

        # put all together
        self.resnet = nn.Sequential(*layers)

        # branches to predict distances and orientations
        self.predict_dist = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, dbins+1, 1)
        )

        self.predict_omega = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, abins+1, 1)
        )
        
        self.predict_theta = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins+1, 1)
        )
        
        self.predict_phi = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins//2+1, 1)
        )

        self.predict_mu = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, 1, 1)
        )
        self.predict_sigma = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, 1, 1),
            nn.ELU()
        )


    def forward(self,msa,xyz_tmp,f1d_tmp,idx=None):

        B, N, L = msa.shape
        if idx == None:
            idx = torch.arange(L, device=msa.device).unsqueeze(0).expand(B,-1)
        
        # features from MSA
        feat_msa = self.msa_feat_extractor(msa,idx)

        # features from templates
        params = {"DMIN":2.0,"DMAX":20.0,"dbins":36}
        #f1d_in = torch.nan_to_num(f1d_tmp)
        #f1d_tmp[torch.isnan(f1d_tmp)] = 0.0
        xyz_in = torch.transpose(xyz_tmp,1,2)
        c6d_in = xyz_to_c6d(xyz_in,params)
        tplt = torch.cat([nn.functional.one_hot(dist_to_bins(c6d_in[...,0],params),37),
                          torch.sin(c6d_in[...,1]).unsqueeze(-1),
                          torch.cos(c6d_in[...,1]).unsqueeze(-1),
                          torch.sin(c6d_in[...,2]).unsqueeze(-1),
                          torch.cos(c6d_in[...,2]).unsqueeze(-1),
                          torch.sin(c6d_in[...,3]).unsqueeze(-1),
                          torch.cos(c6d_in[...,3]).unsqueeze(-1),
                          f1d_tmp.unsqueeze(1).repeat(1,L,1,1),
                          f1d_tmp.unsqueeze(2).repeat(1,1,L,1)], dim=-1)
        tplt = tplt.permute(0,3,1,2).contiguous()
        feat_tplt = self.tplt_feat_extractor(tplt).mean(0,keepdim=True)
        
        # combine features from MSA and templates
        features = torch.cat([feat_msa,feat_tplt.repeat(B,1,1,1)],dim=1)
        features[torch.isnan(features)] = 0.0
        
        if self.gradchk > 1:
            #features.requires_grad = True
            emb = checkpoint_sequential(self.resnet, self.gradchk, features)
        else:
            emb = self.resnet(features)

        #print(torch.isnan(f1d_tmp).sum(), torch.isnan(c6d_in).sum(), torch.isnan(emb).sum())

        embT = torch.transpose(emb,2,3)
        
        # logits for 6D coordinate predictions
        logits = [self.predict_dist(0.5*(emb+embT)),
                  self.predict_omega(0.5*(emb+embT)),
                  self.predict_theta(torch.cat([emb,embT],dim=1)),
                  self.predict_phi(torch.cat([emb,embT],dim=1))]

        # predicted parameters of lognormal distributions
        # for Ca-Ca distances (mu,sigma)
        lognorm = torch.stack([self.predict_mu(0.5*(emb+torch.transpose(emb,2,3))).squeeze(1),
                               self.predict_sigma(0.5*(emb+torch.transpose(emb,2,3))).squeeze(1)+1],dim=-1)

        out = {"dist": logits[0], "omega": logits[1], 
               "theta": logits[2], "phi": logits[3],
               "edge": emb, "lognorm": lognorm}
    
        return out

    
class TRRosetta2_mix(nn.Module):
    """trRosetta base network"""
    
    def __init__(self, 
                 wmin=0.8, dcareg=4.5, gradchk=0,
                 block_type='orig', 
                 dropout=0.15, drates=[1,2,4,8,16], fsize=3, 
                 cycls1=1, cycls2=6, filts1=128, filts2=64, 
                 dbins=36, abins=36, logmix=2 ):
        super(TRRosetta2_mix, self).__init__()
        self.gradchk = gradchk

        if block_type == 'orig':
            self.resblock = ResBlock2D
        elif block_type == 'bottleneck':
            self.resblock = ResBlock2D_bottleneck
        else:
            sys.exit('Wrong residual block type: %s'%(block_type))
            
        self.feat_extractor = MSAFeatureExtractor(wmin=wmin, dcareg=dcareg)

        layers = [
            nn.Conv2d(531, filts1, 1),
            nn.InstanceNorm2d(filts1, affine=True),
            nn.ELU(),
            nn.Dropout(dropout)
        ]

        # first cycle (with more filters)
        for _ in range(cycls1):
            for dilation in drates:
                layers.append(self.resblock(filts1, fsize, dilation, dropout))

        # project down
        layers.append(nn.Conv2d(filts1, filts2, 1))
        layers.append(nn.ELU())

        # second cycle (with less filters)
        for _ in range(cycls2):
            for dilation in drates:
                layers.append(self.resblock(filts2, fsize, dilation, dropout))

        # one more block with dilation 1
        layers.append(self.resblock(filts2, fsize, 1, dropout))

        # put all together
        self.resnet = nn.Sequential(*layers)

        # branches to predict distances and orientations
        self.predict_dist = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, dbins+1, 1)
        )

        self.predict_omega = nn.Sequential(
            self.resblock(filts2, 1, 1, dropout),
            nn.Conv2d(filts2, abins+1, 1)
        )
        
        self.predict_theta = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins+1, 1)
        )
        
        self.predict_phi = nn.Sequential(
            self.resblock(filts2*2, 1, 1, dropout),
            nn.Conv2d(filts2*2, abins//2+1, 1)
        )

        self.predict_mix = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, logmix, 1)
        )
        self.predict_mu = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, logmix, 1)
        )
        self.predict_sigma = nn.Sequential(
            nn.Conv2d(filts2, 128, 1),
            nn.ELU(),
            self.resblock(128, 1, 1, dropout),
            nn.Conv2d(128, logmix, 1),
            nn.ELU()
        )


    def forward(self,msa,idx=None):

        B, N, L = msa.shape
        if idx == None:
            idx = torch.arange(L, device=msa.device).unsqueeze(0).expand(B,-1)
        
        features = self.feat_extractor(msa,idx)
        
        if self.gradchk > 1:
            features.requires_grad = True 
            emb = checkpoint_sequential(self.resnet, self.gradchk, features)
        else:
            emb = self.resnet(features)

        embT = torch.transpose(emb,2,3)
        
        # logits for 6D coordinate predictions
        logits = [self.predict_dist(0.5*(emb+embT)),
                  self.predict_omega(0.5*(emb+embT)),
                  self.predict_theta(torch.cat([emb,embT],dim=1)),
                  self.predict_phi(torch.cat([emb,embT],dim=1))]

        # predicted parameters of lognormal distributions
        # for Ca-Ca distances (mu,sigma)
        lognorm = torch.stack([self.predict_mix(0.5*(emb+torch.transpose(emb,2,3))).permute(0,2,3,1),
                               self.predict_mu(0.5*(emb+torch.transpose(emb,2,3))).permute(0,2,3,1),
                               self.predict_sigma(0.5*(emb+torch.transpose(emb,2,3))).permute(0,2,3,1)+1],dim=-1)

        out = {"dist": logits[0], "omega": logits[1], 
               "theta": logits[2], "phi": logits[3],
               "edge": emb, "lognorm": lognorm}
    
        return out

