# Implement Sergey's simple refinement module in pytorch

import torch
import torch.nn as nn
import torch.nn.functional as F
from resnet import ResidualNetwork

# util functions
def get_dm(xyz, eps=1e-8):
    '''
    Input:
        - xyz: initial CA coordinates or BB coordinates (B, L, 3) or (B, L, 3, 3)
    Output:
        - log(dist_map). diagonal is set to zero. (B, L, L, 1) or (B, 3L, 3L, 1) 
    '''
    B, L = xyz.shape[:2]
    if len(xyz.shape) == 4:
        is_BB = True
        xyz = xyz.reshape(B, -1, 3) # (B,L,3) or (B, 3*L, 3)
    else:
        is_BB = False
    #
    x_x = xyz[:,None,:,:] - xyz[:,:,None,:] # (B, L', L', 3)
    dist = torch.sqrt(torch.square(x_x).sum(dim=-1)+eps) # (B, L', L')
    dist_log = torch.log(dist+eps)
    dist_log = dist_log - torch.tril(torch.triu(dist_log))

    if is_BB: # input is BB coordinates
        dist_log = dist_log.reshape(B,L,3,L,3)
        dist_log = dist_log.permute(0,1,3,2,4).reshape(B,L,L,9)
        return dist_log
    else: # input is CA coordinates
        return dist_log.unsqueeze(-1)

def get_seqsep(idx):
    '''
    Input:
        - idx: residue indices of given sequence (B,L)
    Output:
        - seqsep: sequence separation feature with sign (B, L, L, 1)
                  Sergey found that having sign in seqsep features helps a little
    '''
    seqsep = idx[:,None,:] - idx[:,:,None]
    sign = torch.sign(seqsep)
    seqsep = sign * torch.log(torch.abs(seqsep) + 1.0)
    return seqsep.unsqueeze(-1)

def get_tiled_1d_features(seq1hot, node=None):
    '''
    Input:
        - seq: target sequence in integer (B,L)
    Output:
        - tiled 1d features including 1hot encoded sequence (B, L, 21) and node features if given
    '''
    B,L,_ = seq1hot.shape
    #
    seq1hot = seq1hot.float()
    if node != None:
        feat_1d = torch.cat((seq1hot, node), dim=-1)
    else:
        feat_1d = seq1hot

    left = feat_1d.view(B,L,1,-1).expand(-1,-1,L,-1)
    right = feat_1d.view(B,1,L,-1).expand(-1,L,-1,-1)
    return torch.cat((left, right), dim=-1) # (B, L, L, -1)

class SimpleRefineNet(nn.Module):
    def __init__(self, n_iter=8, n_block=4, n_feat_in=215, n_feat_block=32, n_feat_out=9, block_type='orig', p_drop=0.1):
        super(SimpleRefineNet, self).__init__()
        self.n_iter = n_iter
        # TODO: replace resnet with transformer encoder
        self.resnet = ResidualNetwork(n_block, n_feat_in, n_feat_block, n_feat_block,
                                      block_type=block_type, p_drop=p_drop)
        self.pred_dist = nn.Linear(n_feat_block, n_feat_out)

    def refine_net(self, xyz, dm_diff, eps=1e-8):
        B, L = xyz.shape[:2]
        is_BB = False
        if dm_diff.shape[-1] == 9: # is BB
            is_BB = True
            dm_diff = dm_diff.reshape(B,L,L,3,3).permute(0,1,3,2,4).reshape(B,3*L,3*L,1)
        dm_diff = dm_diff - torch.tril(torch.triu(dm_diff.squeeze(dim=-1))).unsqueeze(dim=-1)

        xyz = xyz.view(B, -1, 3) # (B, L, 3) or (B, 3*L, 3)
        x_x = xyz[:,None,:,:] - xyz[:,:,None,:] # (B,L',L',3)
        dm_orig = torch.sqrt(torch.square(x_x).sum(dim=-1)+eps).unsqueeze(-1) # (B, L', L', 1)
        #g = 2.0 * x_x * (dm_orig - dm_diff) / dm_orig
        g = x_x * (dm_orig - dm_diff) / dm_orig
        g = g.mean(dim=1)
        #
        xyz = xyz - g # 0.5 is learning rate. It could be learnable parameter
        #xyz = xyz - 0.5*g # 0.5 is learning rate. It could be learnable parameter
        if is_BB:
            return xyz.reshape(B,L,3,3)
        else:
            return xyz

    def forward(self, xyz, seq, idx=None, node=None, edge=None):
        # feature generation
        B, L = xyz.shape[:2]
        if idx == None:
            idx = torch.arange(L, device=xyz.device)
            idx = idx.unsqueeze(0).expand(B,-1)
        dm = get_dm(xyz) # distance map from the coordinates
        seqsep = get_seqsep(idx)
        tiled_1d = get_tiled_1d_features(seq, node=node)

        if edge != None:
            feat = torch.cat((edge, dm, seqsep, tiled_1d), dim=-1)
        else:
            feat = torch.cat((dm, seqsep, tiled_1d), dim=-1)
        
        # predict distance map & mask
        feat = feat.permute(0,3,1,2).contiguous()
        feat = self.resnet(feat)
        feat = feat.permute(0,2,3,1).contiguous()

        # Symmetrize features as we're gonna predict symmetric outputs
        feat = 0.5*(feat + feat.permute(0,2,1,3))
        dm_diff = self.pred_dist(feat) # (B, L, L, 9)
        
        out_xyz = list()
        for i_iter in range(self.n_iter):
            xyz = self.refine_net(xyz, dm_diff)
            out_xyz.append(xyz)

        return out_xyz
