#!/usr/bin/env python3

import os,sys,glob
script_dir = os.path.dirname(__file__)
sys.path.insert(0, script_dir)
sys.path.insert(0, script_dir+'/util')

# suppress some pytorch warnings when using a cpu
import warnings
warnings.filterwarnings("ignore", category=UserWarning) 

import json
import time
import importlib
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from args import get_args_score
from parsers import parse_a3m,parse_pdb
from metrics import lDDT,RMSD,KL
from geometry import xyz_to_c6d,c6d_to_bins2
from trfold.trFold import TRFold
from models.ensemble import EnsembleNet
from util import num2aa

C6D_KEYS = ['dist','omega','theta','phi']

def import_method(str):
    '''import a method from folder/subfolder/file.py given a string
    folder.subfolder.file.method; add folder/subfolder to sys.path'''
    module,method = str.rsplit('.',1)
    sys.path.insert(0, script_dir+'/'+module.rsplit('.',1)[0].replace('.','/'))
    module = importlib.import_module(module)
    method = getattr(module,method)
    return method

def write_pdb(filename, bb, seq):
    '''save backbone in PDB format'''
    #atoms = bb.cpu().detach().numpy()
    f = open(filename,"w")
    L = bb.shape[1]
    ctr = 1
    for i in range(L):
        for j,atm_j in enumerate([" N  "," CA "," C  "]):
            f.write ("%-6s%5s %4s %3s %s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n"%(
                "ATOM", ctr, atm_j, num2aa[seq[i]],
                "A", i+1, bb[j,i,0], bb[j,i,1], bb[j,i,2],
                1.0, 0.0 ) )
            ctr += 1
    f.write("TER\n")
    f.close()

def main():
    
    separator = '# ' + '-'*78
    
    ########################################################
    # 0. process inputs
    ########################################################

    # read in info on the registered models and parse arguments
    reg_models = json.load(open(script_dir+"/params/models.json"))
    args = get_args_score(list(reg_models.keys()))
    torch.set_num_threads(args.nthreads)
    print(separator)
    for k,v in reg_models.items():
        flag = '*' if k==args.net else ' '
        print("# %c %-15s   [ens=%d]   %s"%(flag,k,len(v[-1]),v[0]))


    ########################################################
    # 1. load user-selected network
    ########################################################

    sel = reg_models[args.net]
    net_params = json.load(open(sel[2]+'/'+sel[3]))
    chks = [sel[2]+'/'+chk for chk in sel[4]]
    NetClass = import_method(sel[1])
    if len(chks)>1:
        Net = EnsembleNet(NetClass, net_params, chks)
    else:
        Net = NetClass(**net_params)
        weights = torch.load(chks[0],map_location=torch.device(args.device))
        if 'model_state_dict' in weights.keys():
            weights = weights['model_state_dict']
        Net.load_state_dict(weights, strict=False)
        
    Net = Net.to(args.device)
    Net.eval()
    print(separator)
    print("# number of params in %s : %d"%(args.net,sum(p.numel() for p in Net.parameters())))

    
    ########################################################
    # 2. load sequences and structures
    ########################################################

    inputs = []
    if args.list is not None:
        filenames = [line.strip().split() for line in open(args.list).readlines()]
    elif args.pdb_dir is not None:
        filenames = []
        for fn in glob.glob(os.path.join(args.pdb_dir,'*.pdb')):
            fasname = fn.replace('.pdb','.fas')
            if os.path.exists(fasname):
                filenames.append([fasname,fn])
            else:
                filenames.append([fn])

    for fn in filenames:
        if len(fn) == 2: # msa & pdb files both provided
            fas,pdb = fn
            item = [os.path.basename(fas).split('.')[0],
                    os.path.basename(pdb).split('.')[0]]
            fas = parse_a3m(fas)
            pdb = parse_pdb(pdb)
            item += [torch.from_numpy(fas['msa'][None].astype(np.int64)),
                     torch.from_numpy(pdb['xyz'][:,:3].astype(np.float32)).permute(1,0,2),
                     torch.from_numpy(pdb['idx']-1)]
        else: # no msa file, use seq from pdb file
            pdb = fn[0] 
            item = [os.path.basename(pdb).split('.')[0]]*2
            pdb = parse_pdb(pdb)
            item += [torch.from_numpy(pdb['seq'][None,None].astype(np.int64)),
                     torch.from_numpy(pdb['xyz'][:,:3].astype(np.float32)).permute(1,0,2),
                     torch.arange(len(pdb['idx']))] # ignore pdb residue numbering
                     #torch.from_numpy(pdb['idx']-1)]
        inputs.append(item)
        

    ########################################################
    # 3. load background distributions
    ########################################################

    BKGDIR = "/projects/ml/trDesign/backgrounds/generic/%s.npz"
    seq_len = set([i[2].shape[-1] for i in inputs])
    bkg = {l:np.load(BKGDIR%(l)) for l in seq_len}
    bkg_nbytes = [[v2.nbytes for _,v2 in v.items()] for _,v in bkg.items()]
    bkg = {k:[torch.from_numpy(v[k2]).permute(2,0,1) for k2 in C6D_KEYS] for k,v in bkg.items()}
    print("# number of background distributions : %d (%.2fmb)"%(len(seq_len),np.sum(bkg_nbytes)/1024**2))
    

    ########################################################
    # 4. evaluate parsed proteins
    ########################################################

    print(separator)
    print("# %13s %8s %8s %8s %8s | %6s"%('label','KL','CCE','Ca-lDDT','RMSD','time'))
    if args.ocsv is not None: 
        csvfile = open(args.ocsv, 'w')
        print(f'name,kl,cce,lddt,rmsd,time', file=csvfile)

    for i,inp in enumerate(inputs):
        
        tic = time.time()
        
        xyz0 = inp[3].to(args.device)
        idx = inp[4].to(args.device)
        L = inp[2].shape[-1]
        
        with torch.no_grad():
            msa = inp[2].to(args.device)
            out = Net(msa)
        
        logits = [out[key].detach() for key in C6D_KEYS]
        probs = [nn.functional.softmax(l[0], dim=0) for l in logits]

        bkg_i = [p.to(args.device) for p in bkg[L]]
        kl = KL(probs,bkg_i).cpu().detach().numpy()
        
        TRF = TRFold(probs)
        if 'xyz' in out.keys():
            xyz = TRF.fold(ca_trace=out['xyz'][0,:,1], 
                           batch=args.batch,lr=args.lr,nsteps=args.nsteps)
        elif 'lognorm' in out.keys():
            xyz = TRF.fold(lognorm=out['lognorm'],
                           batch=args.batch,lr=args.lr,nsteps=args.nsteps)
        else:
            sys.exit('TRFold cannot generate an initial structure.')
        
        
        c6d = xyz_to_c6d(xyz[None],{'DMAX':20.0})
        c6d_bins = c6d_to_bins2(c6d,{'DMIN':2.0,'DMAX':20.0,'ABINS':36,'DBINS':36})
        c6d_1hot = [F.one_hot(a.squeeze(),n).permute(2,0,1) 
                    for a,n in zip(torch.split(c6d_bins[0],1,dim=-1),(37,37,37,19))]
        
        cce = [(torch.log(p+1e-6)*x-torch.log(q+1e-6)*x).sum(0).mean() for p,q,x in zip(probs,bkg_i,c6d_1hot)]
        cce = torch.stack(cce).mean().cpu().detach().numpy()

        lddt = lDDT(xyz0[1],xyz[1,idx]).cpu().detach().numpy()
        rmsd = RMSD(xyz0[1],xyz[1,idx]).cpu().detach().numpy()
        
        toc = time.time()

        print("%15s %8.5f %8.5f %8.5f %8.4f | %6.3fs"%(inp[0],kl,cce,lddt,rmsd,toc-tic))
        if args.ocsv is not None:
            print(f'{inp[0]},{kl},{cce},{lddt},{rmsd},{toc-tic}', file = csvfile)
        
        if args.opdb is not None:
            write_pdb("%s/%s.pdb"%(args.opdb,inp[0]),
                      xyz.cpu().detach().numpy(),
                      inp[2][0,0])

        if args.onpz is not None:
            c6d = [np.roll(p.permute(1,2,0).cpu().numpy().astype(np.float32),1,axis=-1) for p in probs]
            np.savez_compressed("%s/%s.npz"%(args.onpz,inp[0]),
                                dist=c6d[0],
                                omega=c6d[1],
                                theta=c6d[2],
                                phi=c6d[3])

    if args.ocsv is not None: csvfile.close()
        
    
if __name__ == '__main__':
    main()
